#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#       Copyright (C) 2005-2007 Carabos Coop. V. All rights reserved
#       Copyright (C) 2008-2019 Vicent Mas. All rights reserved
#
#       This program is free software: you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation, either version 3 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#       Author:  Vicent Mas - vmas@vitables.org

"""
Here is defined the BookmarksDlg class.

This module provides a dialog for deleting bookmarks from the bookmarks list.
"""

__docformat__ = 'restructuredtext'

import sys
import os.path

from qtpy import QtCore
from qtpy import QtGui
from qtpy import QtWidgets

translate = QtWidgets.QApplication.translate


class BookmarksDlg(QtWidgets.QDialog):
    """
    The dialog for deleting bookmarks.

    The class defines a modal dialog used to delete entries from the
    bookmarks list. Bookmarks are displayed in a tree view. As a bonus, they
    can be visited directly from this dialog by double clicking them in the
    tree. At the bottom there is a group of buttons made of the ``Delete``,
    ``OK`` and ``Cancel`` buttons.
    ``Delete`` removes checked items from the bookmarks list.

    :Parameters:

    - `blist`: the bookmarks list
    - `hbgui`: an instance of
      :meth:`vitables.docbrowser.browsergui.HelpBrowserGUI` (the parent widget)
    """

    def __init__(self, blist, hbgui):
        """
        Dialog constructor.
        """

        # The HelpBrowser GUI instance from which this dialog has been opened
        super(BookmarksDlg, self).__init__(hbgui)

        self.setWindowTitle(
            translate('BookmarksDlg', 'Bookmarks editor', 'Dialog caption'))
        dlg_layout = QtWidgets.QVBoxLayout(self)

        # Add a tree view
        self.tree = QtWidgets.QTreeView(self)
        self.tree.setItemsExpandable(False)
        self.tmodel = QtGui.QStandardItemModel()
        self.tree.setModel(self.tmodel)
        self.tmodel.setHorizontalHeaderLabels([
            translate('BookmarksDlg', 'Bookmark',
            'First column header of the bookmarks table'),
            translate('BookmarksDlg', 'URL',
            'Second column header of the bookmarks table')])
        dlg_layout.addWidget(self.tree)

        # Add a group of buttons
        self.button_group = QtWidgets.QDialogButtonBox(self)
        self.ok_button = self.button_group.addButton(
            translate('BookmarksDlg', '&OK', 'Button label'),
            QtWidgets.QDialogButtonBox.AcceptRole)
        self.del_button = self.button_group.addButton(
            translate('BookmarksDlg', '&Delete', 'Button label'),
            QtWidgets.QDialogButtonBox.ActionRole)
        self.cancel_button = self.button_group.addButton(
            translate('BookmarksDlg', '&Cancel', 'Button label'),
            QtWidgets.QDialogButtonBox.RejectRole)
        dlg_layout.addWidget(self.button_group)

        # We dont work directly on the HelpBrowser bookmarks list,
        # instead we make a copy. This is convenient if, after doing
        # some changes in the list, we decide to cancel the changes.
        self.blist = blist[:]
        self.fillBookmarksTable()
        self.show()

        # Finally we connect signals to slots
        self.tmodel.itemChanged.connect(self.updateDeleteButton)
        self.tree.doubleClicked.connect(self.displayBookmark)
        self.button_group.clicked.connect(self.buttonClicked)

        self.del_button.setEnabled(False)


    def fillBookmarksTable(self):
        """Add entries to the bookmarks table."""

        self.tmodel.setColumnCount(2)
        parent_item = self.tmodel.invisibleRootItem()
        # Each item of the tree (bookmark) is extracted from the bookmarks list
        for entry in self.blist :
            # extracts the short name. Examples:
            # /home/vmas/estilo.html --> estilo.html
            # /home/vmas/estilo.html#color --> estilo.html#color
            shortname = os.path.basename(entry)
            item = QtGui.QStandardItem(shortname)
            item.setCheckable(True)
            item.setEditable(False)
            item1 = QtGui.QStandardItem(entry)
            item1.setEditable(False)
            parent_item.appendRow([item, item1])
        self.tree.setExpanded(self.tmodel.indexFromItem(parent_item), True)
        self.tree.repaint()
        self.repaint()


    def displayBookmark(self, index):
        """
        Display a given bookmark in the help browser.

        When a bookmark is double clicked the `HelpBrowser` window is
        updated and displays that bookmark.

        :Parameter index: the index in the bookmarks tree view of the selected
          bookmark
        """

        # Get the bookmark UID
        row = self.tmodel.itemFromIndex(index).row()
        src = self.blist[row]
        self.parent().browser.displaySrc(src)


    def updateDeleteButton(self, item):
        """
        Enable/disable the ``Delete`` button.

        The state of the ``Delete`` button depends on the presence/abscence
        of checked items in the bookmarks list. Every time an item is
        clicked the whole bookmarks tree is traversed looking for checked
        items. The``Delete`` button state is updated accordingly.

        :Parameter item: the bookmarks tree item checked/unchecked
        """

        parent_item = self.tmodel.invisibleRootItem()
        # Iterate over the QTreeWidget looking for checked items
        enabled = 0
        row = 0
        item = parent_item.child(row)
        while item != None:
            if item.checkState() == QtCore.Qt.Checked:
                enabled = 1
                break
            row = row + 1
            item = parent_item.child(row)

        self.del_button.setEnabled(enabled)


    def buttonClicked(self, button):
        """
        Action selector method.

        Depending on the button clicked in the button group the
        appropriate method is called.

        :Parameter button: the clicked button
        """

        role = self.button_group.buttonRole(button)
        # OK button clicked
        if role == QtWidgets.QDialogButtonBox.AcceptRole:
            gui = self.parent()
            gui.browser.bookmarks = self.blist
            self.accept()
        # Delete button clicked
        elif role == QtWidgets.QDialogButtonBox.ActionRole:
            self.deleteBookmarks()
        # Cancel button clicked
        elif role == QtWidgets.QDialogButtonBox.RejectRole:
            self.reject()


    def deleteBookmarks(self) :
        """Delete all selected bookmarks."""

        parent_item = self.tmodel.invisibleRootItem()
        # Iterate over the QTreeView looking for checked items
        deleted_rows = []
        row = 0
        item = parent_item.child(row)
        while item != None:
            if item.checkState() == QtCore.Qt.Checked:
                deleted_rows.append(row)
            row = row + 1
            item = parent_item.child(row)
        # Items with highest model indexes are removed first from the model
        # This way deleting an item does not modify the model index of the
        # remaining items
        deleted_rows.reverse()
        for row in deleted_rows:
            self.tmodel.takeRow(row)
            del self.blist[row]

        # After deletion we udpate the dialog
        self.del_button.setEnabled(0)

if __name__ == '__main__':
    APP = QtWidgets.QApplication(sys.argv)
    DLG = BookmarksDlg(['uno', 'dos'], None)
    APP.exec_()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#       Copyright (C) 2005-2007 Carabos Coop. V. All rights reserved
#       Copyright (C) 2008-2019 Vicent Mas. All rights reserved
#
#       This program is free software: you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation, either version 3 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#       Author:  Vicent Mas - vmas@vitables.org

"""This program creates a potentially big table for tests purposes.

The program is based on the tutorial programs included in the PyTables
User's Guide.

Modified to use buffers so that filling times are bettered by one order of
magnitude. F. Altet 2006-02-27
"""

import time
from numpy import arange
import tables

t0 = time.time()

# Describe a particle record
class Particle(tables.IsDescription):
    """Description of a table record."""
    lati        = tables.IntCol(pos=0)
    longi       = tables.IntCol(pos=1)
    temperature = tables.FloatCol(pos=2)
    pressure    = tables.FloatCol(pos=3)

# Open a file in "w"rite mode
fileh = tables.open_file("large_table.h5", mode = "w")

# Get the HDF5 root group
root = fileh.root
# Create a group
group = fileh.create_group(root, "Particles")

# Now, create and fill the tables in Particles group
filters = tables.Filters(complevel=1, complib='lzo', shuffle=1)
nrows = 10**7
table = fileh.create_table("/Particles", "TParticle", Particle, 
    "Sample set of particles ", filters, expectedrows = nrows)

# Number of rows in buffer
nrowsbuf = 10000
# Fill the table with 10**N particles
for i in range(0, nrows, nrowsbuf):
    if i + nrowsbuf > nrows:
        nrapp = nrows - i
    else:
        nrapp = nrowsbuf
    # First, assign the values to the Particle record
    lati = arange(i, i + nrapp)
    longi = 10 - lati
    temperature = lati + 0.5
    pressure = 10.5 - lati
    # This injects the Record values
    table.append([lati, longi, temperature, pressure])

# Flush the table buffers and close the file
fileh.close()

tsec = round((time.time() - t0), 3)
print('Last of the run: {0} s'.format(tsec))



"""This module provides calculator functionality."""

# pylint: disable=W0212

import logging
import os
import re
from vitables.calculator.calculator_dlg import Ui_CalculatorDialog

from qtpy import QtCore
from qtpy import QtGui
from qtpy import QtWidgets
import tables

import numpy as np
import vitables.calculator.evaluator as vtce
import vitables.utils as vtu


translate = QtCore.QCoreApplication.translate

log = logging.getLogger(__name__)


def run():
    gui = vtu.getGui()
    selection_count = len(vtu.getView().selectedIndexes())
    question = ''
    if selection_count == 0:
        question = translate('Calculator', 'No group is selected. '
                             'Relative references are disabled. '
                             'Continue?')
    if selection_count > 1:
        question = translate('Calculator',
                             'Multiple groups are selected. '
                             'Relative references are disabled. '
                             'Continue?')
    if question:
        answer = QtWidgets.QMessageBox.question(
            gui, translate('Calculator', 'Relative references are disabled'),
            question, buttons=QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No)
        if answer != QtWidgets.QMessageBox.Yes:
            return
    dialog = CalculatorDialog()
    dialog.exec_()


# Marker used as prefix to distinguish identifiers from functions.
IDENTIFIER_MARKER = '$'

# Regular expression used to find identifiers in an evaluation formula.
IDENTIFIER_RE = r'\$[\w.]+'


def extract_identifiers(expression):
    """Return a set of identifiers that appear in the expression."""
    return set(re.findall(IDENTIFIER_RE, expression))


def get_current_group():
    """Return selected pytables group.

    If zero or more then one items are selected then None is
    returned. If a group or file is selected then return it. If a
    pytables leaf is selected then return group that contains the
    leaf.

    """
    selected_nodes = vtu.getSelectedNodes()
    if len(selected_nodes) != 1:
        return None
    node = selected_nodes[0]
    if not isinstance(node, tables.Leaf):
        return node
    return node._v_parent


def create_group(group, path):
    """Given ancestor group and a path create a group."""
    if not path:
        return group
    node_name = path[0]
    rest_of_path = path[1:]
    if node_name in group._v_groups:
        return create_group(group._v_groups[node_name], rest_of_path)
    if node_name in group._v_children:  # child but not a group
        return None
    node_group = group._v_file.create_group(group, node_name)
    return create_group(node_group, rest_of_path)


def find_identifier_root(model, identifier):
    """Find the identifier root group in the model.

    Identifier is a string, root group is found by matching item name
    to identifier beggining. The second return argument contains path
    to to identifier node relative to root.

    """
    root_index = QtCore.QModelIndex()
    for row in range(model.rowCount(root_index)):
        index = model.index(row, 0, root_index)
        name = model.data(index, QtCore.Qt.DisplayRole)
        if identifier.startswith(name):
            return (model.nodeFromIndex(index).node,
                    identifier[len(name) + 1:])
    return None, None


def find_node(start_node, path):
    """Given group and a list of node names return node referenced by path."""
    if not path or start_node is None:
        return start_node
    if not isinstance(start_node, tables.Group):
        return None
    if path[0] in start_node._v_children:
        return find_node(start_node._v_children[path[0]], path[1:])
    else:
        return None


def build_identifier_node_dict(identifiers, current_group):
    """Map identifiers to pytables nodes."""
    model = vtu.getModel()
    identifier_node_dict = {}
    for identifier in identifiers:
        identifier_ancestor, relative_path = find_identifier_root(model,
                                                                  identifier)
        if identifier_ancestor is None:
            identifier_ancestor = current_group
            relative_path = identifier
        identifier_node = find_node(identifier_ancestor,
                                    relative_path.split('.'))
        if identifier_node:
            identifier_node_dict[identifier] = identifier_node
    return identifier_node_dict


class CalculatorDialog(QtWidgets.QDialog, Ui_CalculatorDialog):
    def __init__(self, parent=None):
        super(CalculatorDialog, self).__init__(parent)
        self.setupUi(self)
        self._name_expression_dict = {}
        self._settings = QtCore.QSettings()
        self._settings.beginGroup('Calculator')
        self._restore_expressions()

    def on_buttons_rejected(self):
        """Slot for cancel button, save expressions on exit."""
        self._store_expressions()
        self.reject()

    def on_buttons_clicked(self, button):
        """Slot for apply button, run and store saved expressions."""
        button_id = self.buttons.standardButton(button)
        if button_id == QtWidgets.QDialogButtonBox.Apply:
            if self._execute_expression():
                self._store_expressions()
                self.accept()

    @QtCore.Slot()
    def on_save_button_clicked(self):
        """Store expression for future use.

        Ask user for expression name. If provided name is new add it
        to saved expressions list widget. Store in/update name expression
        dictionary.

        """
        current_item = self.saved_list.currentItem()
        if current_item is None:
            # Either the list is empty or nothing is selected.
            current_name = ""
        else:
            current_name = current_item.text()
        name, is_accepted = QtWidgets.QInputDialog.getText(
            self, translate('Calculator', 'Save expression as'),
            translate('Calculator', 'Name:'), text=current_name)
        if not is_accepted:
            return
        if name not in self._name_expression_dict:
            self.saved_list.addItem(name)
        self._name_expression_dict[name] = (self.statements_edit.toPlainText(),
                                            self.expression_edit.toPlainText(),
                                            self.result_edit.text())

    @QtCore.Slot()
    def on_remove_button_clicked(self):
        """Remove stored expression.

        Delete selected row from saved widget and name expression
        dictionary.

        """
        current_row = self.saved_list.currentRow()
        if current_row < 0:
            QtWidgets.QMessageBox.warning(
                self, translate('Calculator', 'Nothing selected'),
                translate('Calculator', 'No saved expression selected '
                          'to be removed.'))
            return
        removed_item = self.saved_list.takeItem(
            self.saved_list.currentRow())
        del self._name_expression_dict[removed_item.text()]

    def on_saved_list_itemSelectionChanged(self):
        """Update expression and result fields from selected item.

        Find first selected widget name, find name in expression
        dictionary and update expression and result widgets.

        """
        selected_indexes = self.saved_list.selectedIndexes()
        if len(selected_indexes) == 0:
            # KFP 2015-08-04: There is a race condition when removing a
            # stored expression.  Both this method and
            # ``on_remove_button_clicked`` are called very close to the
            # same time meaning ``self.saved_list.count()`` can show the
            # wrong value.
            return
        selected_index = selected_indexes[0]
        name = self.saved_list.itemFromIndex(selected_index).text()
        statements, expression, destination = self._name_expression_dict[name]
        self.statements_edit.setText(
            statements if statements is not None else ''
        )
        self.expression_edit.setText(
            expression if expression is not None else ''
        )
        self.result_edit.setText(
            destination if destination is not None else ''
        )

    def _store_expressions(self):
        """Store expressions in configuration.

        Save name expression dictionary in 'expression' part of
        configuration.

        """
        self._settings.beginWriteArray('expressions')
        for index, (name, (statements, expression, destination)) in enumerate(
                self._name_expression_dict.items()):
            self._settings.setArrayIndex(index)
            self._settings.setValue('name', name)
            self._settings.setValue('statements', statements)
            self._settings.setValue('expression', expression)
            self._settings.setValue('destination', destination)
        self._settings.endArray()

    def _restore_expressions(self):
        """Read stored expressions from settings and update list widget.

        Load name expression dictionary from 'expressions' part of
        configuration, update list widget.

        """
        expressions_count = self._settings.beginReadArray('expressions')
        for i in range(expressions_count):
            self._settings.setArrayIndex(i)
            name = self._settings.value('name')
            statements = self._settings.value('statements')
            expression = self._settings.value('expression')
            destination = self._settings.value('destination')
            self._name_expression_dict[name] = (statements, expression,
                                                destination)
            self.saved_list.addItem(name)
        self._settings.endArray()

    def _all_identifiers_found(self, identifiers, identifier_node_dict):
        """Return false if identifier is not found or references a group."""
        for identifier in identifiers:
            if identifier not in identifier_node_dict:
                QtWidgets.QMessageBox.critical(
                    self, translate('Calculator', 'Node not found'),
                    translate('Calculator',
                              'Node "{0}" not found'.format(identifier)))
                return False
            if not isinstance(identifier_node_dict[identifier], tables.Leaf):
                QtWidgets.QMessageBox.critical(
                    self, translate('Calculator', 'Node type'),
                    translate('Calculator',
                              'Node "{0}" does not contain data. '
                              'It is probably a group.'.format(identifier)))
                return False
        return True

    def _create_eval_globals_and_epsression(self, expression,
                                            identifier_node_dict):
        """Transform identifiers into valid python names, update expression."""
        eval_globals = {}
        for index, (identifier, node) in enumerate(
                identifier_node_dict.items()):
            variable_name = (identifier.replace('.', '_')
                             + '_calculator_index_' + str(index))
            expression = expression.replace(
                IDENTIFIER_MARKER + identifier, variable_name)
            eval_globals[variable_name] = node
        return eval_globals, expression

    def _get_result_group_and_name(self):
        """Find or create a group for result based on provided name."""
        result_identifier = self.result_edit.text()
        if not result_identifier:
            QtWidgets.QMessageBox.critical(
                self, translate('Calculator', 'Result name'),
                translate('Calculator',
                          'The location to store results is not specified'))
            return None
        model = vtu.getModel()
        result_ancestor, relative_path = find_identifier_root(
            model, result_identifier)
        if result_ancestor is None:
            result_ancestor = get_current_group()
            relative_path = result_identifier
        relative_path = relative_path.split('.')
        result_group = find_node(result_ancestor, relative_path[:-1])
        if result_group is None:
            answer = QtWidgets.QMessageBox.question(
                self, translate('Calculator', 'Create group'),
                translate('Calculator', 'There is no group "{group}" in '
                          '"{ancestor}". File "{filename}". Create it?'.format(
                              group='/'.join(relative_path[:-1]),
                              ancestor=result_ancestor._v_pathname,
                              filename=result_ancestor._v_file.filename)),
                buttons=QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No)
            if answer != QtWidgets.QMessageBox.Yes:
                return None
            result_group = create_group(result_ancestor, relative_path[:-1])
            if result_group is None:
                QtWidgets.QMessageBox.critical(
                    self, translate('Calculator', 'Result name'),
                    translate('Calculator',
                              'Failed to create group "{group}" inside '
                              '{ancestor} to hold results. File "{filename}".'
                              ''.format(
                                  group='/'.join(relative_path[:-1]),
                                  ancestor=result_ancestor._v_pathname,
                                  filename=result_ancestor._v_file.filename)))
                return None
        result_name = relative_path[-1]
        if result_name in result_group._v_children:
            QtWidgets.QMessageBox.critical(
                self, translate('Calculator', 'Result name'),
                translate('Calculator',
                          'Node "{node}" already exists in group "{group}". '
                          'File "{filename}". Choose another place to store '
                          'results.'.format(
                              node=result_name,
                              group=result_group._v_pathname,
                              filename=result_group._v_file.filename)))
            return None, None
        return result_group, result_name

    def _execute_expression(self):
        """Execute expression and store results.

        Check existence of all tables used in the expression and
        result.

        """
        statements = self.statements_edit.toPlainText()
        expression = self.expression_edit.toPlainText()
        identifier_strings = extract_identifiers(expression)
        identifiers = [i[1:] for i in identifier_strings]
        current_group = get_current_group()
        identifier_node_dict = build_identifier_node_dict(identifiers,
                                                          current_group)
        if not self._all_identifiers_found(identifiers, identifier_node_dict):
            return False
        eval_globals, expression = self._create_eval_globals_and_epsression(
            expression, identifier_node_dict)
        results = self._get_result_group_and_name()
        if results is None:
            return False
        result_group, result_name = results
        try:
            result = vtce.evaluate(statements, expression, eval_globals)
        except Exception as e:
            log.error(str(e))
            QtWidgets.QMessageBox.critical(
                self, translate('Calculator', 'Evaluation error'),
                translate('Calculator', 'An exception was raised during '
                          'evaluation, see log for details.'))
            return False
        if not isinstance(result, np.ndarray):
            if isinstance(result, (list, tuple)):
                result = np.array(result)
            else:
                result = np.array([result])
        try:
            result_group._v_file.create_array(
                result_group, result_name, obj=result,
                title='Expression: ' + self.expression_edit.toPlainText())
        except Exception as e:
            log.error(str(e))
            QtWidgets.QMessageBox.critical(
                self, translate('Calculator', 'Result save error'),
                translate('Calculator', 'An exception was raised while '
                          'trying to store results, see log for details.'))
            return False
        vtu.getModel().updateTreeFromData()
        return True



# -*- coding: utf-8 -*-

#       Copyright (C) 2005, 2006, 2007 Carabos Coop. V. All rights reserved
#       Copyright (C) 2008-2019 Vicent Mas. All rights reserved
#
#       This program is free software: you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation, either version 3 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#       Author:  Vicent Mas - vmas@vitables.org

#
# posterIcons.py
#
import sys

from qtpy import QtGui
from qtpy import QtWidgets

ICONSDIR = '../vitables/icons/'
big_icons = [('file_rw', 'The root node (read-write)'),
    ('file_ro', 'The root node (read-only)'),
    ('dbfilters', 'The root node of the Query Results file  '),
    ('folder', 'A collapsed group '), 
    ('document-open-folder', 'An expanded group ')]
small_icons = [('table', 'A table (heterogeneus dataset) '),
    ('array', 'A regular array (homogeneus dataset) '),
    ('earray', 'An enlargeable array'),
    ('carray', 'A compressed array'),
    ('vlarray', 'A variable length array'),
    ('vlstring', 'An array of VLStrings'),
    ('link_table', 'A soft link to a table (heterogeneus dataset) '),
    ('link_array', 'A soft link to a regular array (homogeneus dataset) '),
    ('link_earray', 'A soft link to an enlargeable array'),
    ('link_carray', 'A soft link to a compressed array'),
    ('link_vlarray', 'A soft link to a variable length array'),
    ('object', 'A serialized objects dataset'),
    ('image-missing', 'An unsupported dataset')]

class Poster(QtWidgets.QMainWindow) :
    def __init__(self) :
        super(QtWidgets.QMainWindow, self).__init__()
        w = QtWidgets.QWidget(self)
        l = QtWidgets.QGridLayout(w)
        self.setCentralWidget(w)
        self.makePoster()


    def makePoster(self):
        widget = self.centralWidget()
        layout = widget.layout()
        r = 0
        c = 0
        iconsDict = {'22x22': big_icons, '16x16': small_icons}
        for key in iconsDict.keys():
            for (name, caption) in iconsDict[key] :
                pixmap = QtGui.QPixmap()
                image_path = '{0}/{1}/{2}.png'.format(ICONSDIR, key, name)
                pixmap.load(image_path)
                imLabel = QtWidgets.QLabel(widget)
                imLabel.setPixmap(pixmap)
                layout.addWidget(imLabel, r, c, 1, 1)
                textLabel = QtWidgets.QLabel(caption,widget)
                layout.addWidget(textLabel, r, c + 1, 1, 1)
                r = r + 1
                if r > 8:
                    r = 0
                    c = 2


def main(args) :
    app = QtWidgets.QApplication(args)
    poster = Poster()
    poster.show()
    app.exec_()

if __name__ == '__main__' :
    main(sys.argv)




#       Copyright (C) 2005-2007 Carabos Coop. V. All rights reserved
#       Copyright (C) 2008-2019 Vicent Mas. All rights reserved
#
#       This program is free software: you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation, either version 3 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#       Author:  Vicent Mas - vmas@vitables.org

__version__ = '3.0.2'
__all__ = ["docbrowser", "h5db", "logger", "nodeprops", "plugins",
           "queries", "preferences", "vttables", "vtwidgets",
           "utils", "vtsite", "vtapp", "vtgui", "vtsplash"]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#       Copyright (C) 2005-2007 Carabos Coop. V. All rights reserved
#       Copyright (C) 2008-2019 Vicent Mas. All rights reserved
#
#       This program is free software: you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation, either version 3 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#       Author:  Vicent Mas - vmas@vitables.org

#
#       This script is based on a set of scripts by Francesc Alted.

"""Sample of a typical object tree."""

import tables

class Particle(tables.IsDescription):
    identity = tables.StringCol(itemsize=22, dflt=" ", pos=0)  # character String
    idnumber = tables.Int16Col(dflt=1, pos = 1)  # short integer
    speed    = tables.Float32Col(dflt=1, pos = 1)  # single-precision

# Open a file in "w"rite mode
fileh = tables.open_file("objecttree.h5", mode = "w")
# Get the HDF5 root group
root = fileh.root

# Create the groups:
group1 = fileh.create_group(root, "group1")
group2 = fileh.create_group(root, "group2")

# Now, create an array in root group
array1 = fileh.create_array(root, "array1",
    [bytes("string", encoding="utf-8"), bytes("array", encoding="utf-8")], "Data array")
# Create 2 new tables in group1
table1 = fileh.create_table(group1, "table1", Particle)
table2 = fileh.create_table("/group2", "table2", Particle)
# Create the last table in group2
array2 = fileh.create_array("/group1", "array2", [1,2,3,4])

# Now, fill the tables:
for table in (table1, table2):
    # Get the record object associated with the table:
    row = table.row
    # Fill the table with 10 records
    for i in range(10):
        # First, assign the values to the Particle record
        row['identity']  = 'This is particle: %2d' % (i)
        row['idnumber'] = i
        row['speed']  = i * 2.
        # This injects the Record values
        row.append()

    # Flush the table buffers
    table.flush()

# Finally, close the file (this also will flush all the remaining buffers!)
fileh.close()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#       Copyright (C) 2005-2007 Carabos Coop. V. All rights reserved
#       Copyright (C) 2008-2019 Vicent Mas. All rights reserved
#
#       This program is free software: you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation, either version 3 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#       Author:  Vicent Mas - vmas@vitables.org

"This creates an HDF5 file with a potentially large number of objects"

import sys
import numpy
import tables

filename = 'large_number_of_objects.h5'

# Open a new empty HDF5 file
fileh = tables.open_file(filename, mode = "w")

# nlevels -- Number of levels in hierarchy
# ngroups -- Number of groups on each level
# ndatasets -- Number of arrays on each group
# LR: Low ratio groups/datasets
#nlevels, ngroups, ndatasets = (3, 1, 1000)
# MR: Medium ratio groups/datasets
nlevels, ngroups, ndatasets = (3, 10, 100)
# HR: High ratio groups/datasets
#nlevels, ngroups, ndatasets = (30, 10, 10)

# Create an Array to save on disk
a = numpy.array([-1, 2, 4], dtype=numpy.int16)

group = fileh.root
group2 = fileh.root
for k in range(nlevels):
    for j in range(ngroups):
        for i in range(ndatasets):
            # Save the array on the HDF5 file
            fileh.create_array(group2, 'array'+str(i), a, "Signed short array")
        # Create a new group
        group2 = fileh.create_group(group, 'group'+str(j))
    # Create a new group
    group3 = fileh.create_group(group, 'ngroup'+str(k))
    # Iterate over this new group (group3)
    group = group3
    group2 = group3

fileh.close()




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#       Copyright (C) 2005-2007 Carabos Coop. V. All rights reserved
#       Copyright (C) 2008-2019 Vicent Mas. All rights reserved
#
#       This program is free software: you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation, either version 3 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#       Author:  Vicent Mas - vmas@vitables.org

"""
This module provides a dialog for solving node renaming issues on the tree of
databases view.

Some times naming problems appear when the tree is being edited:

  - when a file is being saved as a different one and its name is
    already being used by another file in the same directory
  - when a new group is being created and its name is already being
    used by other node in the destination group
  - when a node is being renamed and its new name is already being
    used by other node in the same group
  - when a node is being pasted/dropped and its name is already being
    used by other node in the destination group

This dialog allows the user to solve the problem. Available options are:

  - rename the object in trouble
  - overwrite the existing object
  - cancel the editing action

"""

import os.path
import re

from qtpy import QtCore
from qtpy import QtGui
from qtpy import QtWidgets

from qtpy.uic import loadUiType

import vitables.utils

__docformat__ = 'restructuredtext'

translate = QtWidgets.QApplication.translate
# This method of the PyQt5.uic module allows for dinamically loading user
# interfaces created by QtDesigner. See the PyQt5 Reference Guide for more
# info.
Ui_RenameNodeDialog = \
    loadUiType(os.path.join(os.path.dirname(__file__), 'rename_dlg.ui'))[0]


class RenameDlg(QtWidgets.QDialog, Ui_RenameNodeDialog):
    """
    Ask user for help when a name issue raises.

    By loading UI files at runtime we can:

        - create user interfaces at runtime (without using pyuic)
        - use multiple inheritance, MyParentClass(BaseClass, FormClass)


    Regular Qt class `QInputDialog` is not used because, at least apparently,
    it doesn't provide a way for customizing buttons text.

    This dialog allows the user to solve naming issues. Available options
    are:

      - rename the object in trouble
      - overwrite the existing object
      - cancel the editing action

    :Parameters:

    - `name`: the troubled name
    - `pattern`: a regular expression pattern that the name must match
    - `info`: dialog title and label
    """

    def __init__(self, name, pattern, info):
        """A customised `QInputDialog`.
        """

        vtapp = vitables.utils.getVTApp()
        # Makes the dialog and gives it a layout
        super(RenameDlg, self).__init__(vtapp.gui)
        self.setupUi(self)

        # Sets dialog caption
        self.setWindowTitle(info[0])
        self.generalInfo.setText(info[1])

        self.troubled_name = name
        self.cpattern = re.compile(pattern)

        # The result returned by this dialog will be contained in the
        # action dictionary. Its values can be:
        # If Overwrite --> True, troubled_name
        # If Rename --> False, new_name
        # If Cancel --> False, ''
        self.action = {'overwrite': False, 'new_name': ''}

        # The dialog buttons: Rename, Overwrite and Cancel
        self.overwrite_button = self.buttonsBox.addButton(
            translate('RenameDlg', 'Overwrite', 'A button label'),
            QtWidgets.QDialogButtonBox.AcceptRole)
        self.rename_button = self.buttonsBox.addButton(
            translate('RenameDlg', 'Rename', 'A button label'),
            QtWidgets.QDialogButtonBox.AcceptRole)
        self.rename_button.setDefault(1)
        self.cancel_button = self.buttonsBox.button(
            QtWidgets.QDialogButtonBox.Cancel)

        # Setup a validator for checking the entered node name
        validator = QtGui.QRegExpValidator(self)
        qt_pattern = QtCore.QRegExp(pattern)
        validator.setRegExp(qt_pattern)
        self.valueLE.setValidator(validator)
        self.valueLE.setText(self.troubled_name)

        self.valueLE.selectAll()

        # Make sure that buttons are in the proper activation state
        self.valueLE.textChanged.emit(self.valueLE.text())

    @QtCore.Slot('QString', name="on_valueLE_textChanged")
    def checkName(self, new_name):
        """
        Check the new name value.

        Every time that the text box content changes, this method is
        asked to check if the new name and the original name differ.
        Four cases can occur:

            1) the new name is empty
            2) the new name and the current nodename are the same
            3) the new name and the current name differ and the new name
               is not being used by a sibling node
            4) the new name and the current name differ and the new name
               is being used by a sibling node

        In cases 1) and 2) the new name is not valid, the `Rename` and
        `Overwrite` buttons are both disabled.
        In case 3) the `Rename` button is enabled and the `Overwrite` one is
        not. In case 4) the `Overwrite` button is enabled but the `Rename` one
        is not.

        Beware that case 2) can appear only during a node renaming operation or
        during a file `Save As...` operation (not when creating a new group or
        pasting/dropping a node)

        :Parameter new_name: the value currently displayed in the text box
        """

        # If the new name and the current nodename are the same then the
        # group 1 of the pattern is matched
        result = self.cpattern.search(new_name)
        if (result is None) or \
                ((result is not None) and (result.lastindex == 1)) or \
                (not new_name):
            self.rename_button.setEnabled(0)
            self.overwrite_button.setEnabled(0)
        elif new_name == self.troubled_name:
            self.rename_button.setEnabled(0)
            if self.troubled_name == '/':
                self.overwrite_button.setEnabled(0)
            else:
                self.overwrite_button.setEnabled(1)
        else:
            self.rename_button.setEnabled(1)
            self.overwrite_button.setEnabled(0)

    @QtCore.Slot("QAbstractButton *", name="on_buttonsBox_clicked")
    def executeAction(self, button):
        """Execute the action specified by the clicked button.

        :Parameter button: the clicked button (`Rename` or `Overwrite`)
        """

        if button == self.rename_button:
            self.rename_node()
        elif button == self.overwrite_button:
            self.overwriteNode()

    def overwriteNode(self):
        """
        Set the new name and exit.

        When the `Overwrite` button is clicked the new name is saved and
        the dialog finishes.
        """

        self.action['overwrite'] = True
        self.action['new_name'] = self.troubled_name
        self.accept()

    def rename_node(self):
        """
        Set the new name and exit.

        When the `Rename` button is clicked the new name is saved and
        the dialog finishes.
        """

        new_name = self.valueLE.text()
        self.action['overwrite'] = False
        self.action['new_name'] = new_name
        self.accept()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#       Copyright (C) 2005-2007 Carabos Coop. V. All rights reserved
#       Copyright (C) 2008-2019 Vicent Mas. All rights reserved
#
#       This program is free software: you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation, either version 3 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#       Author:  Vicent Mas - vmas@vitables.org

"""
This module implements a controller for managing the queries.

The manager tracks the existing filtered tables and their names, launches the
Query dialog and executes the queries at low level (i.e. `PyTables` level).
It also keeps a description of the last executed query and tracks the tables
currently being queried, in order to ensure that no more than 1 query at a time
is executed on a given table.
"""

__docformat__ = 'restructuredtext'

import logging
import vitables.utils

from qtpy import QtCore
from qtpy import QtGui
from qtpy import QtWidgets

import vitables.queries.query as query
import vitables.queries.querydlg as querydlg


translate = QtWidgets.QApplication.translate

log = logging.getLogger(__name__)


def getTableInfo(table):
    """Retrieves table info required for querying it.

    :Parameter table: the `tables.Table` instance being queried.
    """
    info = {}
    info['nrows'] = table.nrows
    info['src_filepath'] = table._v_file.filename
    info['src_path'] = table._v_pathname
    info['name'] = table._v_name
    # Fields info: top level fields names, flat fields shapes and types
    info['col_names'] = frozenset(table.colnames)
    info['col_shapes'] = \
        dict((k, v.shape) for (k, v) in table.coldescrs.items())
    info['col_types'] = table.coltypes
    # Fields that can be queried
    info['condvars'] = {}
    info['valid_fields'] = []

    if info['nrows'] <= 0:
        log.info(
            translate('QueriesManager',
                      "Table {0} is empty. Nothing to query.",
                      'Info message for users').format(info['name']))
        return None

    # Find out the valid (i.e. searchable) fields and condition variables.
    # First discard nested fields.
    # Beware that order matters in binary operations that mix set instances
    # with frozensets: set & frozenset returns a set but frozenset & set
    # returns a frozenset
    valid_fields = set(info['col_shapes'].keys()).intersection(
        info['col_names'])
    # info['col_names'].intersection(info['col_shapes'].keys())

    # Then discard fields that aren't scalar and those that are complex
    for name in valid_fields.copy():
        if (info['col_shapes'][name] != ()) or \
           info['col_types'][name].count('complex'):
            valid_fields.remove(name)

    # Among the remaining fields, those whose names contain blanks
    # cannot be used in conditions unless they are mapped to
    # variables with valid names
    index = 0
    for name in valid_fields.copy():
        if name.count(' '):
            while ('col{0}'.format(index)) in valid_fields:
                index = index + 1
            info['condvars']['col{0}'.format(index)] = \
                table.cols._f_col(name)
            valid_fields.remove(name)
            valid_fields.add('col{0} ({1})'.format(index, name))
            index = index + 1
    info['valid_fields'] = valid_fields

    # If table has not columns suitable to be filtered does nothing
    if not info['valid_fields']:
        log.info(
            translate('QueriesManager',
                      """Table {0} has no columns suitable to be """
                      """queried. All columns are nested, multidimensional """
                      """or have a Complex data type.""",
                      'Info when trying to query a table').format(
                          info['name']))
        return None
    elif len(info['valid_fields']) != len(info['col_names']):
        # Log a message if non selectable fields exist
        log.warning(
            translate('QueriesManager',
                      """Some table columns contain multidimensional, """
                      """nested or Complex data. They cannot be queried so """
                      """are not included in the Column selector of the """
                      """query dialog.""", 'An informational note for users'))

    return info


class QueriesManager(QtCore.QObject):
    """This is the class in charge of the execution of queries.

    `PyTables` doesn't support threaded queries. So when several queries are
    requested to ``ViTables`` they will be executed sequentially.

    Also no more than one query can be made at the same time on a given table.
    This goal is achieved in a very simple way: tracking the tables currently
    being queried in a data structure (a dictionary at present).

    :Parameter parent: the parent of the `QueriesManager` object
    """

    def __init__(self, parent=None):
        """Setup the queries manager.

        The last query description has three components: the filepath of
        the file where the queried table lives, the nodepath of the queried
        table and the query condition.

        A query name is the name of the table where the query results are
        stored. By default it has the format "Filtered_TableUID" where UID
        is an integer automatically generated. User can customise the query
        name in the New Query dialog.
        """

        super(QueriesManager, self).__init__(parent)

        # Description of the last query made
        self.last_query = [None, None, None]
        # UID for automatically generating query names
        self.counter = 0
        # The list of query names currently in use
        self.ft_names = []

        self.vtapp = vitables.utils.getVTApp()
        self.vtgui = self.vtapp.gui
        self.dbt_view = self.vtgui.dbs_tree_view
        self.dbt_model = self.vtgui.dbs_tree_model

    def newQuery(self):
        """Process the query requests launched by users.
        """

        # The VTApp.updateQueryActions method ensures that the current node is
        # tied to a tables.Table instance so we can query it without
        # further checking
        current = self.dbt_view.currentIndex()
        node = self.dbt_model.nodeFromIndex(current)
        table_uid = node.as_record
        table = node.node

        table_info = getTableInfo(table)
        if table_info is None:
            return

        # Update the suggested name sufix
        self.counter = self.counter + 1
        query_description = self.getQueryInfo(table_info, table)
        if query_description is None:
            self.counter = self.counter - 1
            return

        # Update the list of names in use for filtered tables
        self.ft_names.append(query_description['ft_name'])
        self.last_query = [query_description['src_filepath'],
                           query_description['src_path'],
                           query_description['condition']]

        # Run the query
        tmp_h5file = self.dbt_model.tmp_dbdoc.h5file
        new_query = query.Query(tmp_h5file, table_uid, table,
                                query_description)
        new_query.query_completed.connect(self.addQueryResult)
        QtWidgets.qApp.setOverrideCursor(QtCore.Qt.WaitCursor)
        new_query.run()

    def getQueryInfo(self, info, table):
        """Retrieves useful info about the query.

        :Parameters:

        - `info`: dictionary with info about the queried table
        - `table`: the tables.Table instance being queried
        """

        # Information about table
        # Setup the initial condition
        last_query = self.last_query
        if (last_query[0], last_query[1]) == (info['src_filepath'],
                                              info['src_path']):
            initial_condition = last_query[2]
        else:
            initial_condition = ''

        # GET THE QUERY COMPONENTS
        # Get a complete query description from user input: condition to
        # be applied, involved range of rows, name of the
        # filtered table and name of the column of returned indices
        query_dlg = querydlg.QueryDlg(info, self.ft_names,
                                      self.counter, initial_condition, table)
        try:
            query_dlg.exec_()
        finally:
            query_description = dict(query_dlg.query_info)
            del query_dlg
            QtWidgets.qApp.processEvents()

        if not query_description['condition']:
            return None

        # SET THE TITLE OF THE RESULT TABLE
        title = query_description['condition']
        for name in info['valid_fields']:
            # Valid fields can have the format 'fieldname' or
            # 'varname (name with blanks)' so a single blank shouldn't
            # be used as separator
            components = name.split(' (')
            if len(components) > 1:
                fieldname = '({0}'.format(components[-1])
                title = title.replace(components[0], fieldname)
        query_description['title'] = title

        return query_description

    def deleteAllQueries(self):
        """
        Delete all nodes under the `Query results` node of the databases tree.
        """

        title = translate('QueriesManager', 'Cleaning the Query results file',
                          'Caption of the QueryDeleteAll dialog')
        text = translate('QueriesManager', ('\n\nYou are about to delete all '
                                            'nodes under Query results\n\n'),
                         'Ask for confirmation')
        itext = ''
        dtext = ''
        buttons = {
            'Delete': (translate('QueriesManager', 'Delete', 'Button text'),
                       QtWidgets.QMessageBox.YesRole),
            'Cancel': (translate('QueriesManager', 'Cancel', 'Button text'),
                       QtWidgets.QMessageBox.NoRole)}
        # Ask for confirmation
        answer = vitables.utils.questionBox(title, text, itext, dtext, buttons)
        if answer == 'Cancel':
            return

        # Remove every filtered table from the tree of databases model/view
        model_rows = self.dbt_model.rowCount(QtCore.QModelIndex())
        tmp_index = self.dbt_model.index(model_rows - 1, 0,
                                         QtCore.QModelIndex())
        rows_range = list(
            range(0, self.dbt_model.rowCount(tmp_index)).__reversed__())
        for row in rows_range:
            index = self.dbt_model.index(row, 0, tmp_index)
            self.vtapp.nodeDelete(index, force=True)

        # Reset the queries manager
        self.counter = 0
        self.ft_names = []

    def addQueryResult(self, completed, table_uid):
        """Update the GUI once the query has finished.

        Add the result of the query to the tree of databases view and open
        the new filtered table.

        :Parameters:

        - `completed`: whether the query has been succesful or not
        - `table_uid`: the UID of the table just queried
        """

        QtWidgets.qApp.restoreOverrideCursor()
        if not completed:
            log.error(translate('QueriesManager',
                                'Query on table {0} failed!',
                                'Warning log message about a failed '
                                'query').format(table_uid))
            return

        # Update temporary database view i.e. call lazyAddChildren
        model_rows = self.dbt_model.rowCount(QtCore.QModelIndex())
        tmp_index = self.dbt_model.index(model_rows - 1, 0,
                                         QtCore.QModelIndex())
        self.dbt_model.lazyAddChildren(tmp_index)

        # The new filtered table is inserted in first position under
        # the Query results node and opened
        index = self.dbt_model.index(0, 0, tmp_index)
        self.vtapp.nodeOpen(index)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#       Copyright (C) 2005-2007 Carabos Coop. V. All rights reserved
#       Copyright (C) 2008-2019 Vicent Mas. All rights reserved
#
#       This program is free software: you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation, either version 3 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#       Author:  Vicent Mas - vmas@vitables.org

"""
This module provides a dialog for querying (filtering) `tables.Table` nodes.
The result of the query is stored in other `tables.Table` node, referred as a
filtered table, which will live in the temporary database (labeled as `Query
results` in the databases tree).
"""

import logging
import os.path
import vitables.utils

import numpy
from qtpy import QtCore
from qtpy import QtGui
from qtpy import QtWidgets
from qtpy.uic import loadUiType


__docformat__ = 'restructuredtext'

translate = QtWidgets.QApplication.translate

# This method of the PyQt5.uic module allows for dinamically loading user
# interfaces created by QtDesigner. See the PyQt5 Reference Guide for more
# info.
Ui_QueryDialog = \
    loadUiType(os.path.join(os.path.dirname(__file__), 'query_dlg.ui'))[0]

log = logging.getLogger(__name__)


class QueryDlg(QtWidgets.QDialog, Ui_QueryDialog):
    """
    A dialog for filtering `tables.Table` nodes.

    By loading UI files at runtime we can:

        - create user interfaces at runtime (without using pyuic)
        - use multiple inheritance, MyParentClass(BaseClass, FormClass)

    The dialog layout looks like this::

        root
            label + textbox
            checkbox + label + textbox
            text box
            combobox + combobox + combobox
            label + textbox
            label + textbox
            label + textbox
            Ok + Cancel

    A text box is used to enter the condition to be applied. Combos
    are used to enter operators, functions and field names. Textboxes
    are used to enter the range of the query.

    :Parameters:

    - `info`: a dictionary with information about the table being queried
    - `ft_names`: the list of filtered tables names currently in use
    - `counter`: a counter used to give a unique ID to the query
    - `initial_query`: the dialog will be setup with this query (if any)
    - `table`: the table being queried
    """

    def __init__(self, info, ft_names, counter, initial_query, table):
        """
        Initialise the dialog.
        """
        #
        # Attributes used by slot updateOKState
        #
        self.used_names = ft_names
        self.col_names = info['col_names']
        self.condvars = info['condvars']
        self.source_table = table
        self.num_rows = info['nrows']

        #
        # Attributes used by slot composeQuery
        #
        # If the dialog is cancelled these initial values will be returned to
        # the caller. If the dialog is accepted then these values will be
        # updated (in the slotAccept method) and returned to the caller
        self.query_info = {}
        self.query_info['condition'] = ''
        self.query_info['rows_range'] = ()
        self.query_info['ft_name'] = ''
        self.query_info['indices_field_name'] = ''
        self.query_info['condvars'] = self.condvars
        self.query_info['src_filepath'] = info['src_filepath']
        self.query_info['src_path'] = info['src_path']

        #
        # Create the dialog and customise the content of some widgets
        #
        super(QueryDlg, self).__init__(QtWidgets.qApp.activeWindow())
        self.setupUi(self)

        self.setWindowTitle(translate('QueryDlg', 'New query on table: {0}',
                                      'A dialog caption').format(info['name']))

        self.nameLE.setText('FilteredTable_{0}'.format(counter))

        self.indicesColumnLE.setEnabled(0)

        # Use the values of the last query done on this table (if any)
        if initial_query != '':
            self.queryLE.setText(initial_query)

        # Fill the combos
        operators = ['&', '|', '~', '<', '<=', '==', '!=', '>', '>=',
                     '+', '-', '*', '/', '**', '%']
        self.operatorsComboBox.insertItems(0, operators)
        functions = ['where', 'sin', 'cos', 'tan', 'arcsin', 'arccos',
                     'arctan', 'arctan2', 'sinh', 'cosh', 'tanh',
                     'arcsinh', 'arccosh', 'arctanh', 'log', 'log10', 'log1p',
                     'exp', 'expm1', 'sqrt', 'real', 'imag', 'complex']
        self.functionsComboBox.insertItems(0, functions)
        sorted_fields = [field for field in info['valid_fields']]
        sorted_fields.sort()
        self.columnsComboBox.insertItems(0, sorted_fields)
        self.rstartLE.setText('0')
        self.rstopLE.setText('{0}'.format(info['nrows']))

        whatsthis_button = self.buttonBox.button(
            QtWidgets.QDialogButtonBox.Help)
        whatsthis_button.setText("&What's this")

        #
        # Setup a validator for Range selectors section
        #
        validator = QtGui.QRegExpValidator(QtCore.QRegExp("\\d*"), self)
        self.rstartLE.setValidator(validator)
        self.rstopLE.setValidator(validator)
        self.rstep.setValidator(validator)

        #
        # Connect signals to slots
        #
        self.buttonBox.helpRequested.connect(
            QtWidgets.QWhatsThis.enterWhatsThisMode)
        # Ensure that if the condition line edit is initialised with an
        # initial condition then the OK button will be enabled
        self.nameLE.textChanged.emit(self.nameLE.text())

    @QtCore.Slot(bool, name="on_indicesCheckBox_toggled")
    def enableIndicesColumn(self, cb_on):
        """
        Enable/disable the indices column name field.

        If the checkbox in the global options groupbox is checked then
        a column with the indices of the filtered rows will be added
        to the table produced by the filtering process. Moreover the
        user can set the name of this column.

        :Parameter cb_on: a boolean that indicates if the checkbox is down
            or not.
        """

        if cb_on:
            self.indicesColumnLE.setEnabled(1)
        else:
            self.indicesColumnLE.setEnabled(0)

    @QtCore.Slot("QString", name="on_operatorsComboBox_activated")
    def insertOperator(self, operator):
        """
        Insert an operator in the query line editor.

        :Parameter operator: is the operator in the combobox current item
        """
        self.queryLE.insert(' {0} '.format(operator))

    @QtCore.Slot("QString", name="on_columnsComboBox_activated")
    def insertField(self, field_id):
        """
        Insert a fieldname in the query line editor.

        When the fields combobox is activated the selected field_id is
        inserted in the query line editor. Fields whose names contain
        blanks have a field id with the format 'condvar (fieldname)'.
        In this case only the condvar name will be inserted in the query
        line editor.

        :Parameter field_id: is the field identifier in the combobox current
            item
        """
        self.queryLE.insert(field_id.split(' ')[0])

    @QtCore.Slot("QString", name="on_functionsComboBox_activated")
    def insertFunction(self, text):
        """
        Insert a function in the query line editor.

        When the functions combobox is activated the selected value is
        inserted in the query line editor.

        :Parameter text: is the text of the combobox current item
        """

        name2call = {'where': 'where(B, N, N)',
                     'sin': 'sin(F|C)',
                     'cos': 'cos(F|C)',
                     'tan': 'tan(F|C)',
                     'arcsin': 'arcsin(F|C)',
                     'arccos': 'arccos(F|C)',
                     'arctan': 'arctan(F|C)',
                     'arctan2': 'arctan2(F, F)',
                     'sinh': 'sinh(F|C)',
                     'cosh': 'cosh(F|C)',
                     'tanh': 'tanh(F|C)',
                     'arcsinh': 'arcsinh(F|C)',
                     'arccosh': 'arccosh(F|C)',
                     'arctanh': 'arctanh(F|C)',
                     'log': 'log(F|C)',
                     'log10': 'log10(F|C)',
                     'log1p': 'log1p(F|C)',
                     'exp': 'exp(F|C)',
                     'expm1': 'expm1(F|C)',
                     'sqrt': 'sqrt(F|C)',
                     'real': 'real(C)',
                     'imag': 'imag(C)',
                     'complex': 'complex(F, F)'}
        self.queryLE.insert(name2call[text])

    @QtCore.Slot("QString", name="on_nameLE_textChanged")
    @QtCore.Slot("QString", name="on_indicesColumnLE_textChanged")
    @QtCore.Slot("QString", name="on_queryLE_textChanged")
    @QtCore.Slot("QString", name="on_rstartLE_textChanged")
    @QtCore.Slot("QString", name="on_rstopLE_textChanged")
    def updateOKState(self):
        """
        Update the activation state of the `OK` button.

        Every time that the table name text box changes the new
        content is checked. If it is empty, then the `OK` button is
        disabled.
        If the indices checkbox is checked, every time that the indices column
        name text box changes the
        new content is checked. If it is empty, then the `OK` button
        is disabled.
        Every time that the lower/upper limit text box change the
        new content is checked. If the new values don't make a legal
        condition, then the `OK` button is disabled.
        Every time the column name combobox changes the datatype
        of limit values is checked. If there is a mismatch, then the
        `OK` button is disabled.
        Every time that the spinboxes content change the new content
        is checked. If the combination of the start/stop/step boxes
        makes a not legal selection range, then the `OK` button is
        disabled.
        """

        status_ok = True

        # Check the table name
        ft_name = self.nameLE.text()
        if not ft_name:
            status_ok = False
        elif ft_name in self.used_names:
            status_ok = False
            log.error(
                translate('QueryDlg',
                          """The chosen name is already in use. Please,"""
                          """ choose another one.""",
                          'A logger info message'))

        # Check the indices column name
        indices_colname = self.indicesColumnLE.text()
        if self.indicesColumnLE.isEnabled():
            if indices_colname == '':
                status_ok = False
                log.error(
                    translate('QueryDlg',
                              "Enter a name for the column of indices, please",
                              'A logger info message'))
            elif indices_colname.count('/'):
                status_ok = False
                log.error(
                    translate('QueryDlg',
                              """The chosen name for the column of indices"""
                              """is not valid. It contains '/' characters""",
                              'A logger info message'))
            elif indices_colname in self.col_names:
                status_ok = False
                log.error(
                    translate('QueryDlg',
                              """The chosen name for the column of indices"""
                              """ is already in use. """
                              """"Please, choose another one.""",
                              'A logger info message'))

        # Check the condition.
        # If it is an empty string the OK button is not enabled
        condition = self.queryLE.text()
        if condition.isspace() or (condition in [None, '']):
            status_ok = False

        # Check the range values
        start_str = self.rstartLE.text()
        stop_str = self.rstopLE.text()
        if not (start_str and stop_str):
            status_ok = False
        else:
            start = numpy.array(start_str).astype(numpy.int64)
            stop = numpy.array(stop_str).astype(numpy.int64)
            if stop > self.num_rows:
                status_ok = False
                log.error(
                    translate('QueryDlg',
                              """The stop value cannot be greater than """
                              """the number of rows.""",
                              'A logger info message'))
            elif start > stop:
                status_ok = False
                log.error(
                    translate('QueryDlg',
                              """The start value cannot be greater than the """
                              """stop value.""",
                              'A logger info message'))

        # Enable/disable the OK button
        ok_button = self.buttonBox.button(QtWidgets.QDialogButtonBox.Ok)
        if status_ok:
            ok_button.setEnabled(1)
        else:
            ok_button.setEnabled(0)

    def checkConditionSyntax(self, condition):
        """Check the condition syntax.

        :Parameter condition: the query condition used for filtering the table
        """

        syntax_ok = True
        try:
            self.source_table.will_query_use_indexing(condition, self.condvars)
        except SyntaxError as error:
            syntax_ok = False
            log.error(error.__doc__)
            vitables.utils.formatExceptionInfo()
        except NameError as error:
            syntax_ok = False
            log.error(error.__doc__)
            vitables.utils.formatExceptionInfo()
        except ValueError as error:
            syntax_ok = False
            log.error(error.__doc__)
            vitables.utils.formatExceptionInfo()
        except TypeError as error:
            syntax_ok = False
            log.error(error.__doc__)
            vitables.utils.formatExceptionInfo()
        except:
            syntax_ok = False
            vitables.utils.formatExceptionInfo()
        return syntax_ok

    @QtCore.Slot(name="on_buttonBox_accepted")
    def composeQuery(self):
        """Slot for composing the query and accept the dialog."""

        # Get the query
        condition = self.queryLE.text()
        # Blanks at the begining of condition raise a SyntaxError
        condition = condition.strip()
        if not self.checkConditionSyntax(condition):
            return

        self.query_info['condition'] = condition

        # Get the table name and the name of the column with indices (if any)
        self.query_info['ft_name'] = self.nameLE.text()
        if self.indicesColumnLE.isEnabled():
            self.query_info['indices_field_name'] = \
                self.indicesColumnLE.text()

        # Get the range and convert it into a Python range
        self.query_info['rows_range'] = (
            numpy.array(self.rstartLE.text()).astype(numpy.int64),
            numpy.array(self.rstopLE.text()).astype(numpy.int64),
            numpy.array(self.rstep.text()).astype(numpy.int64))

        # Exit
        self.accept()
